import random

from elo_viz.constants import RESULT_DRAW, RESULT_WIN, RESULT_LOSS
from elo_viz.player import Player


class Match:
    def __init__(self, match_id: str, player1: Player, player2: Player):
        self.id = match_id
        self.player1 = player1
        self.player2 = player2
        self.winner = None
        self.is_draw = False

    def pick_random_winner(self):
        random_number = random.randint(1, 100)
        if random_number <= 33:
            self.winner = self.player1.id
        elif 33 < random_number <= 66:
            self.is_draw = True
        else:
            self.winner = self.player2.id
        
        self.update_rating()

    def update_result(self, winner: int = None, is_draw: bool = False):
        if is_draw is True:
            self.is_draw = is_draw
        else:
            self.winner = winner

        self.update_rating()

    def update_rating(self):
        if self.is_draw is True:
            player1_result = RESULT_DRAW
            player2_result = RESULT_DRAW
        elif self.winner == self.player1.id:
            player1_result = RESULT_WIN
            player2_result = RESULT_LOSS
        elif self.winner == self.player2.id:
            player1_result = RESULT_LOSS
            player2_result = RESULT_WIN

        self.player1.update_rating(self.player2, result=player1_result)
        self.player2.update_rating(self.player1, result=player2_result)

    def display_result(self):
        if self.is_draw is True:
            print(self.player1.name, ' drew with ', self.player2.name)
        elif self.winner == self.player1.id:
            print(self.player1.name, ' won against ', self.player2.name)
        elif self.winner == self.player2.id:
            print(self.player2.name, ' won against ', self.player1.name)

    def to_dict(self):
        return {
            'id': self.id, 'player1': self.player1.to_dict(), 'player2': self.player2.to_dict(),
            'winner': self.winner, 'is_draw': self.is_draw
        }



from threading import Lock, Thread


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """

    _instances = {}

    _lock: Lock = Lock()
    """
    We now have a lock object that will be used to synchronize threads during
    first access to the Singleton.
    """

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        with cls._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class EloConstant(metaclass=SingletonMeta):
    # Singleton class for fun, and to practice using singletons
    K: int = 50

    def __init__(self, K: int) -> None:
        self.K = K


RESULT_WIN = "win"
RESULT_LOSS = "loss"
RESULT_DRAW = "draw"

DEFAULT_RATING = 1000

WIN = 1.0
LOSS = 0.0
DRAW = 0.5

RESULT_TO_SCORE = {
    RESULT_WIN: WIN,
    RESULT_DRAW: DRAW,
    RESULT_LOSS: LOSS
}

# Top 100 Chess Players as of Feb, 2025
CHESS_PLAYERS = [
    ['Carlsen, Magnus', 'NOR', '2833'],
    ['Caruana, Fabiano', 'USA', '2803'],
    ['Nakamura, Hikaru', 'USA', '2802'],
    ['Erigaisi Arjun', 'IND', '2801'],
    ['Gukesh D', 'IND', '2777'],
    ['Abdusattorov, Nodirbek', 'UZB', '2766'],
    ['Firouzja, Alireza', 'FRA', '2760'],
    ['Wei, Yi', 'CHN', '2755'],
    ['Nepomniachtchi, Ian', 'RUS', '2754'],
    ['Anand, Viswanathan', 'IND', '2750'],
    ['So, Wesley', 'USA', '2747'],
    ['Aronian, Levon', 'USA', '2745'],
    ['Dominguez Perez, Leinier', 'USA', '2741'],
    ['Praggnanandhaa R', 'IND', '2741'],
    ['Duda, Jan-Krzysztof', 'POL', '2739'],
    ['Le, Quang Liem', 'VIE', '2739'],
    ['Ding, Liren', 'CHN', '2734'],
    ['Niemann, Hans Moke', 'USA', '2734'],
    ['Keymer, Vincent', 'GER', '2731'],
    ['Mamedyarov, Shakhriyar', 'AZE', '2731'],
    ['Aravindh, Chithambaram VR.', 'IND', '2729'],
    ['Vachier-Lagrave, Maxime', 'FRA', '2729'],
    ['Giri, Anish', 'NED', '2728'],
    ['Fedoseev, Vladimir', 'SLO', '2724'],
    ['Vidit, Santosh Gujrathi', 'IND', '2721'],
    ['Rapport, Richard', 'HUN', '2718'],
    ['Topalov, Veselin', 'BUL', '2717'],
    ['Yu, Yangyi', 'CHN', '2715'],
    ['Dubov, Daniil', 'RUS', '2701'],
    ['Wang, Hao', 'CHN', '2701'],
    ['Sindarov, Javokhir', 'UZB', '2700'],
    ['Radjabov, Teimour', 'AZE', '2698'],
    ['Svidler, Peter', 'FID', '2696'],
    ['Esipenko, Andrey', 'FID', '2695'],
    ['Andreikin, Dmitry', 'FID', '2695'],
    ['Harikrishna, Pentala', 'IND', '2695'],
    ['Deac, Bogdan-Daniel', 'ROU', '2692'],
    ['Sevian, Samuel', 'USA', '2692'],
    ['Artemiev, Vladislav', 'RUS', '2691'],
    ['Liang, Awonder', 'USA', '2690'],
    ['Robson, Ray', 'USA', '2689'],
    ['Van Foreest, Jorden', 'NED', '2688'],
    ['Grischuk, Alexander', 'RUS', '2687'],
    ['Nihal Sarin', 'IND', '2687'],
    ['Bu, Xiangzhi', 'CHN', '2684'],
    ['Kasimdzhanov, Rustam', 'UZB', '2681'],
    ['Maghsoodloo, Parham', 'IRI', '2680'],
    ['Navara, David', 'CZE', '2677'],
    ['Sarana, Alexey', 'SRB', '2672'],
    ['Shankland, Sam', 'USA', '2670'],
    ['Vitiugov, Nikita', 'ENG', '2670'],
    ['Vokhidov, Shamsiddin', 'UZB', '2670'],
    ['Anton Guijarro, David', 'ESP', '2668'],
    ['Howell, David W L', 'ENG', '2668'],
    ['Nguyen, Thai Dai Van', 'CZE', '2668'],
    ['Sadhwani, Raunak', 'IND', '2666'],
    ['Tabatabaei, M. Amin', 'IRI', '2666'],
    ['Leko, Peter', 'HUN', '2666'],
    ['Svane, Frederik', 'GER', '2665'],
    ['Adams, Michael', 'ENG', '2664'],
    ['Saric, Ivan', 'CRO', '2661'],
    ['Alekseenko, Kirill', 'AUT', '2661'],
    ['Oparin, Grigoriy', 'USA', '2660'],
    ['Sargsyan, Shant', 'ARM', '2660'],
    ['Gledura, Benjamin', 'HUN', '2659'],
    ['Yakubboev, Nodirbek', 'UZB', '2659'],
    ['Eljanov, Pavel', 'UKR', '2658'],
    ['Christiansen, Johan-Sebastian', 'NOR', '2658'],
    ['Wojtaszek, Radoslaw', 'POL', '2658'],
    ['Murzin, Volodar', 'FID', '2657'],
    ['Mamedov, Rauf', 'AZE', '2656'],
    ['Morozevich, Alexander', 'RUS', '2656'],
    ['Amin, Bassem', 'EGY', '2654'],
    ['Jones, Gawain C B', 'ENG', '2654'],
    ['Yuffa, Daniil', 'ESP', '2654'],
    ['Shevchenko, Kirill', 'ROU', '2653'],
    ['Dardha, Daniel', 'BEL', '2652'],
    ['Karthikeyan, Murali', 'IND', '2651'],
    ['Indjic, Aleksandar', 'SRB', '2650'],
    ['Inarkiev, Ernesto', 'RUS', '2650'],
    ['Gelfand, Boris', 'ISR', '2649'],
    ['Shirov, Alexei', 'ESP', '2648'],
    ['Cheparinov, Ivan', 'BUL', '2646'],
    ['Ma, Qun', 'CHN', '2645'],
    ['Martirosyan, Haik M.', 'ARM', '2645'],
    ['Grandelius, Nils', 'SWE', '2644'],
    ['Bluebaum, Matthias', 'GER', '2643'],
    ['Kollars, Dmitrij', 'GER', '2643'],
    ['Sjugirov, Sanan', 'HUN', '2643'],
    ['Volokitin, Andrei', 'UKR', '2643'],
    ['Malakhov, Vladimir', 'FID', '2642'],
    ['Mendonca, Leon Luke', 'IND', '2641'],
    ['Bjerre, Jonas Buhl', 'DEN', '2640'],
    ['Bacrot, Etienne', 'FRA', '2640'],
    ['Wang, Yue', 'CHN', '2640'],
    ['Warmerdam, Max', 'NED', '2638'],
    ['Vallejo Pons, Francisco', 'ESP', '2638'],
    ['Najer, Evgeniy', 'FID', '2637'],
    ['Melkumyan, Hrant', 'ARM', '2636'],
    ['Puranik, Abhimanyu', 'IND', '2636'],
    ['Xiong, Jeffery', 'USA', '2636']
]



from .constants import (
    DEFAULT_RATING, RESULT_WIN, RESULT_LOSS, RESULT_DRAW
)
from .matches import Match
from .player import Player
from .rating import Rating, calculate_rating
from .elo import Elo, EloChess



from elo_viz.constants import RESULT_DRAW, RESULT_WIN, RESULT_LOSS, DEFAULT_RATING
from elo_viz.rating import Rating, calculate_rating


class Player:
    def __init__(self, id: str, name: str, rating: int = DEFAULT_RATING, variability: float = 1.0):
        self.id = id
        self.name = name
        self.rating = Rating(rating)
        self.variability = variability
        self.matches = 0
        self.wins = 0
        self.losses = 0
        self.draws = 0

    def update_rating(self, opponent, result):

        if result not in [RESULT_WIN, RESULT_LOSS, RESULT_DRAW]:
            raise ValueError("Incorrect value for result parameter!")

        self.matches += 1

        self.rating = calculate_rating(self.rating, opponent.rating, result)
        self.wins += 1 if result == RESULT_WIN else 0
        self.losses += 1 if result == RESULT_LOSS else 0
        self.draws += 1 if result == RESULT_DRAW else 0

    def to_dict(self):
        return {
            'id': self.id, 'name': self.name, 'rating': float(self.rating),
            'matches': self.matches, 'wins': self.wins, 'losses': self.losses, 'draws': self.draws
        }

    def __str__(self):
        if self.matches:
            win_percentage = self.wins / self.matches * 100
        else:
            win_percentage = 0
        return "Player: {name} Rating: {rating} Win %: {win_percentage}".format(
            name=self.name, rating=self.rating, win_percentage=win_percentage
        )

    def __eq__(self, other):
        return self.id == other.id



import uuid
from random import randint, choice
from typing import List
from elo_viz.constants import DEFAULT_RATING, CHESS_PLAYERS
from elo_viz.player import Player
from elo_viz.rating import Rating
from elo_viz.matches import Match

class Elo:

    def __init__(self, players: List[Player], verbose=False):
        self.players = players
        self.matches = []
        self.verbose = verbose

    def add_player(self, player: Player):
        self.players.append(player)

    def random_match(self):
        p1 = choice(self.players)
        p2 = p1
        while p1 == p2:
            p2 = choice(self.players)
        m = self.match(p1, p2)
        m.display_result()
        self.matches.append(m.to_dict())

    def match(self, p1: Player, p2: Player):
        match = Match(match_id=uuid.uuid4().hex, player1=p1, player2=p2)
        match.pick_random_winner()
        return match

    def print_ranks(self):
        print()
        s = sorted(self.players, key=lambda a: a.rating, reverse=True)
        for i, player in enumerate(s):
            print(i+1, player)


class EloChess:

    def __init__(self, players):
        if players > 100:
            players = 100
        self.players = [Player(uuid.uuid4().hex, CHESS_PLAYERS[i][0], CHESS_PLAYERS[i][2]) for i in range(players)]
        self.matches = []
        self.output_match = True

    def random_match(self):
        p1 = choice(self.players)
        p2 = p1
        while p1 == p2:
            p2 = choice(self.players)
        m = self.match(p1, p2)
        m.display_result()
        self.matches.append(m.to_dict())

    def match(self, p1: Player, p2: Player):
        match = Match(match_id=uuid.uuid4().hex, player1=p1, player2=p2)
        match.pick_random_winner()
        return match

    def print_ranks(self):
        print()
        s = sorted(self.players, key=lambda a: a.rating, reverse=True)
        for i, player in enumerate(s):
            print(i+1, player)


def main():
    e = EloChess(30)
    fast = True
    e.output_match = not fast
    matches = 0
    total_matches_to_simulate = 100
    while 1:
        matches += 1
        e.random_match()
        if fast and (matches % 100000):
            continue
        e.print_ranks()
        if matches > total_matches_to_simulate:
            break


if __name__ == '__main__':
    main()



from elo_viz.constants import EloConstant, DEFAULT_RATING, RESULT_TO_SCORE


class Rating:
    def __init__(self, value=None):
        if value is None:
            value = DEFAULT_RATING
        self.value = float(value)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __eq__(self, other):
        return float(self) == float(other)

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __add__(self, other):
        self.value += other
        return self

    def __sub__(self, other):
        self.value -= other
        return self

    def __str__(self):
        return "Rating: {}".format(str(self.value))

    # def __repr__(self):
    #     return "Rating(value={})".format(str(self.value))


def calculate_expected_score(rating1: Rating, rating2: Rating) -> float:
    expected_score = 1 / (1 + 10 ** ((float(rating2) - float(rating1)) / 400))
    return expected_score


def calculate_rating_from_expected_score(rating: Rating, expected_score: float, result: str) -> Rating:
    score = RESULT_TO_SCORE[result]
    rating += EloConstant.K * (score - expected_score)
    return rating


def calculate_rating(rating1: Rating, rating2: Rating, result: str) -> Rating:
    expected_score = calculate_expected_score(rating1, rating2)
    rating1 = calculate_rating_from_expected_score(
        rating1, expected_score, result)
    return rating1
